"""Module implementing a base for standard ONNX operators, which use the functionality of ONNX node-level inference."""
from typing import TYPE_CHECKING, Callable, Dict, Tuple

import numpy
import onnx
import onnx.reference
import onnx.shape_inference
from onnx.defs import OpSchema

from . import _value_prop
from ._exceptions import InferenceError
from ._node import Node
from ._schemas import SCHEMAS
from ._scope import Scope
from ._shape import SimpleShape
from ._type_system import Optional, Sequence, Tensor, Type
from ._utils import from_array
from ._value_prop import PropValueType

if TYPE_CHECKING:
    from ._graph import Graph


class StandardNode(Node):
    """
    Base type for a Node which has a known reference Schema (``self.schema``), extracted based on the ``op_type``.
    Implements ``infer_output_types_onnx``, which is used for type inference.
    This is based on the functionality of ``onnx.shape_inference.infer_node_outputs``.
    The default ``infer_output_types`` is overriden to use this function.
    """

    @classmethod
    def get_schema(cls) -> OpSchema:
        return SCHEMAS[cls.op_type.domain][cls.op_type.version][cls.op_type.identifier]

    @property
    def schema(self) -> OpSchema:
        return self.get_schema()

    @property
    def min_input(self) -> int:
        return self.schema.min_input

    @property
    def min_output(self) -> int:
        return self.schema.min_output

    def to_singleton_onnx_model(
        self, *, dummy_outputs: bool = True, with_dummy_subgraphs: bool = True
    ) -> Tuple[onnx.ModelProto, Scope]:
        """
        Build a singleton model consisting of just this StandardNode. Used for type inference.
        Dummy subgraphs are typed, but have no graph body, so that we can avoid the build cost.
        They refer to non-existent nodes, but ONNX does not raise an error (for now?).
        """
        # Prepare names for the values in scope of the node
        scope = Scope()
        scope.node[self] = "_this_"
        for key, var in self.inputs.get_vars().items():
            if var not in scope.var:
                scope.var[var] = key
        for key, var in self.outputs.get_vars().items():
            if var not in scope.var:
                scope.var[var] = key
        # We inject the evaluated attribute values here and then substitute back
        self_attrs = self.attrs
        try:
            current_fields = self_attrs.get_fields().items()
            self.attrs = self.Attributes(
                **{k: v.deref() if v is not None else None for k, v in current_fields}
            )
            node_proto: onnx.NodeProto
            # Subgraphs are not fully built for possibly significant performance gains.
            # However, this uses a trick so that they type correctly.
            # This may throw if we are building ``not with_subgraphs``.
            build_subgraph = _make_dummy_subgraph if with_dummy_subgraphs else None
            (node_proto,) = self.to_onnx(scope, build_subgraph=build_subgraph)
        finally:
            self.attrs = self_attrs
        # Create a singleton graph for type inference with our node
        # Input types
        input_info = [
            var.unwrap_type()._to_onnx_value_info(key)
            for key, var in self.inputs.get_vars().items()
        ]

        # Output types with placeholder empty TypeProto (or actual type if not using dummies)
        def out_value_info(curr_key, curr_var):
            if dummy_outputs or curr_var.type is None or not curr_var.type._is_concrete:
                return onnx.helper.make_value_info(curr_key, onnx.TypeProto())
            return curr_var.unwrap_type()._to_onnx_value_info(curr_key)

        output_info = [
            out_value_info(key, var) for key, var in self.outputs.get_vars().items()
        ]
        # Initializers, passed in to allow partial data propagation
        #  - used so that operators like Reshape are aware of constant shapes
        initializers = [
            from_array(var._value.value, key)
            for key, var in self.inputs.get_vars().items()
            if var._value and isinstance(var._value.value, numpy.ndarray)
        ]
        #  Graph and model
        graph = onnx.helper.make_graph(
            [node_proto],
            "StandardOpNode_infer_output_types_onnx",
            input_info,
            output_info,
            initializers,
        )
        # Subgraph internals are hidden by make_dummy_subgraph - so we don't care about subgraphs' opset requirements
        model = onnx.helper.make_model(
            graph,
            opset_imports=[
                onnx.helper.make_operatorsetid(
                    self.op_type.domain, self.op_type.version
                )
            ],
        )
        return model, scope

    def infer_output_types_onnx(self) -> Dict[str, Type]:
        """Execute type & shape inference with ``onnx.shape_inference.infer_node_outputs``."""
        # Check that all (specified) inputs have known types, as otherwise we fail
        if any(var.type is None for var in self.inputs.get_vars().values()):
            return {}

        model, _ = self.to_singleton_onnx_model()

        # Attempt to do shape inference - if an error is caught, we extend the traceback a bit
        try:
            typed_model = onnx.shape_inference.infer_shapes(
                model, check_type=True, strict_mode=True, data_prop=True
            )
        except InferenceError as e:
            raise type(e)(
                f"{str(e)} -- for {self.schema.name}: {self.signature}"
            ) from e

        # Recover the types from protobuf, ignoring empty protobuf objects for failing/unimplemented type inference.
        results = {
            info.name: Type._from_onnx(info.type)
            for info in typed_model.graph.output
            if info.type != onnx.TypeProto()
        }
        # Strips some unuseful type data (unknown dimensions become global-scoped dimension parameters).
        return {
            key: _strip_dim_symbol(type_, lambda x: x.startswith("unk__"))
            for key, type_ in results.items()
        }

    def propagate_values_onnx(self) -> Dict[str, PropValueType]:
        """Perform value propagation by evaluating singleton model.

        The backend used for the propagation can be configured with the `spox._standard.ValuePropBackend` variable.
        """
        # Cannot do propagation when some inputs were not propagated/inferred
        if any(
            var.type is None or var._value is None
            for var in self.inputs.get_vars().values()
        ):
            return {}
        if next(iter(self.subgraphs), None) is not None:
            # Cannot do propagation with subgraphs implicitly for performance - should be reimplemented
            return {}
        model, scope = self.to_singleton_onnx_model(with_dummy_subgraphs=False)
        wrap_feed, run, unwrap_feed = _value_prop.get_backend_calls()
        input_feed = {
            scope.var[var]: wrap_feed(var._value)
            for var in self.inputs.get_vars().values()
            if var._value
        }

        output_feed = run(model, input_feed)

        results = {
            scope.var[str(name)]._which_output: unwrap_feed(
                scope.var[str(name)].unwrap_type(), result
            ).value
            for name, result in output_feed.items()
        }
        return {k: v for k, v in results.items() if k is not None}

    def infer_output_types(self) -> Dict[str, Type]:
        return self.infer_output_types_onnx()

    def propagate_values(self) -> Dict[str, PropValueType]:
        if _value_prop._VALUE_PROP_BACKEND != _value_prop.ValuePropBackend.NONE:
            return self.propagate_values_onnx()
        return {}


def _strip_dim_symbol_shape(
    shape: SimpleShape, pred: Callable[[str], bool]
) -> SimpleShape:
    """
    Remove all instances all ``unk__`` dimension parameters from a shape -- created when running
    ONNX shape inference and no parameter to use existed. It is stripped to avoid conflicts (due to global scoping).
    """
    if shape is None:
        return shape
    xs = [None if isinstance(x, str) and pred(x) else x for x in shape]
    return tuple(xs)


def _strip_dim_symbol(typ: Type, pred: Callable[[str], bool]) -> Type:
    """Apply ``_strip_unk_param_shape`` to all shapes present in this ``Type``."""
    if isinstance(typ, Tensor):
        return Tensor(typ.dtype, _strip_dim_symbol_shape(typ.shape, pred))
    elif isinstance(typ, Sequence):
        return Sequence(_strip_dim_symbol(typ.elem_type, pred))
    elif isinstance(typ, Optional):
        return Optional(_strip_dim_symbol(typ.elem_type, pred))
    else:
        return typ


def _make_dummy_subgraph(_node: Node, key: str, graph: "Graph") -> onnx.GraphProto:
    """
    Make a dummy GraphProto that has inputs and outputs typed like Graph, without a graph body.

    Assumes ``graph`` requests arguments to avoid building it. The body is avoided by having the outputs be equal to
    unknown forwarded values from the outer scope.

    - inputs: ``-> __dummy_input{i}``
    - outputs: ``<- dummy_output{i} <- __dummy_outer_output{i} <- ?``

    Warning: this method may break if ONNX starts throwing errors against missing value names in shape inference.
    """
    if graph.requested_arguments is None:
        raise RuntimeError("Subgraph must have requested_arguments to construct dummy.")

    inputs = []
    for i, arr in enumerate(graph.requested_arguments):
        inputs.append(arr.unwrap_type()._to_onnx_value_info(f"__dummy_input{i}"))

    value_infos = []
    nodes = []
    outputs = []
    for i, arr in enumerate(graph.requested_results.values()):
        outer = f"__dummy_outer_output{i}"
        value_infos.append(arr.unwrap_type()._to_onnx_value_info(outer))
        out = f"__dummy_output{i}"
        outputs.append(arr.unwrap_type()._to_onnx_value_info(out))
        nodes.append(onnx.helper.make_node("Identity", [outer], [out]))
    return onnx.helper.make_graph(
        nodes, f"__dummy_{key}", inputs, outputs, value_info=value_infos
    )



import numpy
import pytest

from spox._shape import Constant, Natural, Shape, ShapeError, Unknown
from spox._type_system import Tensor


@pytest.fixture(params=[numpy.float64, numpy.int32, numpy.bool_])
def scalar_type(request):
    return request.param


@pytest.fixture(
    params=[
        (Shape((Constant(2), Constant(3))), (2, 3)),
        (Shape((Unknown("x"), Unknown("y"))), ("x", "y")),
        (Shape((Unknown(), Unknown("x"), Constant(4))), (None, "x", 4)),
        (Shape(None), None),
    ]
)
def shape_pair(request):
    return request.param


def test_from_onnx_shape_elem(shape_pair):
    if shape_pair[0] is None or shape_pair[1] is None:
        return
    for x, y in zip(shape_pair[0], shape_pair[1]):
        assert x == Natural.from_simple(y)


def test_to_onnx_shape_elem(shape_pair):
    if shape_pair[0] is None or shape_pair[1] is None:
        return
    for x, y in zip(shape_pair[0], shape_pair[1]):
        assert x.to_simple() == y


def test_from_onnx_shape(shape_pair):
    assert shape_pair[0] == Shape.from_simple(shape_pair[1])


def test_to_onnx_shape(shape_pair):
    assert shape_pair[0].to_simple() == shape_pair[1]


def test_shape_rank():
    assert Shape.from_simple((1, 2, 3)).rank == 3
    assert Shape.from_simple((1, 2, 3)).maybe_rank == 3
    assert Shape(None).maybe_rank is None
    with pytest.raises(ShapeError):
        _ = Shape(None).rank


def test_shape_indexing():
    assert Shape.from_simple(("N", "M"))[1].to_simple() == "M"
    assert Shape.from_simple((1, 2, 3))[2].to_simple() == 3
    with pytest.raises(ShapeError):
        _ = Shape(None)[1]
    assert Shape.from_simple(("X", 2, 3, "Y"))[1:3] == Shape.from_simple((2, 3))


def test_tensor_shorthand(scalar_type, shape_pair):
    rich, simple = shape_pair
    assert Tensor(scalar_type, simple)._shape == rich


def test_bad_tensor_type():
    assert Tensor(numpy.float32) != Tensor(numpy.float64)  # sanity check
    for typ in (numpy.object_, object, None):
        with pytest.raises(TypeError):
            Tensor(typ)



import numpy
import pytest

from spox._graph import arguments, results
from spox._type_system import Tensor


@pytest.fixture(scope="session")
def bracket_graph(ext):
    return results(pairs=ext.match_brackets(*arguments(xs=Tensor(numpy.int64, ("N",)))))


@pytest.mark.parametrize(
    "brackets,result",
    [
        ([], []),
        ([1, -1], [(0, 1)]),
        ([1, 1, 1, -1, -1, 1, -1, -1], [(2, 3), (1, 4), (5, 6), (0, 7)]),
        ([1, 1, 0, -1, 0, -1], [(1, 3), (0, 5)]),
        ([1], None),
        ([-1], None),
        ([1, 1, -1], None),
        ([1, -1, -1], None),
    ],
)
def test_sequence_bracket_matching(onnx_helper, bracket_graph, brackets, result):
    def get_bracket_matching(seq):
        ret = onnx_helper.run(bracket_graph, "pairs", xs=numpy.array(seq, dtype=int))
        return [(x, y) for x, y in ret] if ret is not None else None

    assert get_bracket_matching(brackets) == result



import numpy as np
import pytest

import spox.opset.ai.onnx.v17 as op
from spox._attributes import AttrString, AttrStrings, AttrTensor


def test_bad_attribute_type_cast_fails():
    with pytest.raises(TypeError):
        # to must be an value which can be handled by `np.dtype`.
        op.cast(op.const(1), to="abc")


def test_cast_with_build_in_type():
    op.cast(op.const(1), to=str)


def test_float_instead_of_int_attr():
    with pytest.raises(TypeError):
        op.concat([op.const(1)], axis=3.14)  # type: ignore


@pytest.mark.parametrize(
    "vals, expected, field",
    [
        ([3.14], [3.14], "double_data"),
        (["foo"], [b"foo"], "string_data"),
        # Check scalar values
        (3.14, [3.14], "double_data"),
        ("foo", [b"foo"], "string_data"),
    ],
)
def test_tensor_does_not_use_raw_data(vals, field, expected):
    attr = AttrTensor(np.array(vals), "foo")
    pb = attr._to_onnx_deref()
    assert pb.t.raw_data == b""
    assert getattr(pb.t, field) == expected
    assert pb.t.dims == list(np.array(vals).shape)


def test_non_ascii_string_attr():
    AttrString("🐍", "foo")._to_onnx()
    AttrStrings(["🐍"], "foo")._to_onnx()



"""Module implementing the main public interface functions in Spox."""

import contextlib
import itertools
from typing import Dict, List, Optional, Protocol

import numpy as np
import onnx
from onnx.numpy_helper import to_array

from . import _internal_op
from ._attributes import AttrType
from ._graph import Argument, initializer, results
from ._inline import _Inline
from ._standard import _strip_dim_symbol
from ._type_system import Type
from ._var import Var


def argument(typ: Type) -> Var:
    """
    Create an argument variable which may be used as a model input.

    Parameters
    ----------
    typ
        The type of the created argument variable.

    Returns
    -------
    arg
        An unnamed argument variable of given type that may be used as
        a model input to build a graph.
    """
    return _internal_op.Argument(
        _internal_op.Argument.Attributes(type=AttrType(typ, "dummy"), default=None)
    ).outputs.arg


@contextlib.contextmanager
def _temporary_renames(**kwargs: Var):
    # The build code can't really special-case variable names that are
    # not just ``Var._name``.  So we set names here and reset them
    # afterwards.
    name: Optional[str]
    pre: Dict[Var, Optional[str]] = {}
    try:
        for name, arg in kwargs.items():
            pre[arg] = arg._name
            arg._rename(name)
        yield
    finally:
        for arg, name in pre.items():
            arg._rename(name)


def build(inputs: Dict[str, Var], outputs: Dict[str, Var]) -> onnx.ModelProto:
    """
    Builds an ONNX Model with given model inputs and outputs.

    Additional data such as docstrings and metadata can be added to
    the returned ``onnx.ModelProto`` using tools from the
    ``onnx.helper`` module.

    Parameters
    ----------
    inputs
        Model inputs. Keys are names, values must be results of
        ``argument``.
    outputs
        Model outputs. Keys are names, values may be any ``Var``.
        Building will resolve what nodes were used in the construction
        of output variables.

    Returns
    -------
    onnx.ModelProto
        An ONNX ModelProto containing operators necessary to compute
        ``outputs`` from ``inputs``.  If multiple versions of the
        ``ai.onnx`` domain are present, the nodes are all converted to
        the newest one. The minimum ``ai.onnx`` version is set to 14 to
        avoid tooling issues with legacy versions.

    Examples
    --------
    >>> import numpy as np
    >>> import onnxruntime
    >>> from spox import argument, build, Tensor
    >>> import spox.opset.ai.onnx.v17 as op
    >>> # Construct a Tensor type representing a (1D) vector of float32, of size N.
    >>> VectorFloat32 = Tensor(np.float32, ('N',))
    >>> # a, b, c are all vectors and named the same in the graph
    >>> # We create 3 distinct arguments
    >>> a, b, c = [argument(VectorFloat32) for _ in range(3)]
    >>> # p represents the Var equivalent to a * b
    >>> q = op.add(op.mul(a, b), c)
    >>> # Build an ONNX model in Spox
    >>> model = build({'a': a, 'b': b, 'c': c}, {'r': q})
    """
    if not all(isinstance(var, Var) for var in inputs.values()):
        seen_types = {type(obj) for obj in inputs.values()}
        raise TypeError(f"Build inputs must be Vars, not {seen_types - {Var}}.")
    if not all(isinstance(var, Var) for var in outputs.values()):
        seen_types = {type(obj) for obj in outputs.values()}
        raise TypeError(f"Build outputs must be Vars, not {seen_types - {Var}}.")
    if not all(isinstance(var._op, Argument) for var in inputs.values()):
        raise TypeError(
            "Build inputs must be `Var`s constructed using the `spox.argument` function. "
            "They must not be results of other operations."
        )
    if not outputs:
        raise ValueError("Build outputs must not be empty for the graph to be valid.")

    with _temporary_renames(**inputs):
        graph = results(**outputs)
        graph = graph.with_arguments(*inputs.values())
        return graph.to_onnx_model()


class _InlineCall(Protocol):
    """
    A callable returned by ``inline``, taking positional and keyword
    arguments of type ``Var``, and returning a dictionary of names
    (``str``) into ``Var``.
    """

    def __call__(self, *args: Var, **kwargs: Var) -> Dict[str, Var]:
        """
        Parameters
        ----------
        args
            Variables passed as model inputs - positional, as they are
            listed in the model.
        kwargs
            Further variables passed as model inputs - keyword, as
            they are named in the model.
        Returns
        -------
        Dict[str, Var]
            Variables representing the inlined model's outputs.
        """


def _copy_model(model: onnx.ModelProto) -> onnx.ModelProto:
    copied = onnx.ModelProto()
    copied.CopyFrom(model)
    return copied


def inline(model: onnx.ModelProto) -> _InlineCall:
    """Inline an existing ONNX model, taking and producing ``Var``.

    Any valid model may be inlined. The behaviour of the ``model`` is
    replicated, its metadata (docstring, annotations) may be stripped.
    The opset imports of the target model are significant and the
    model itself may be adapted if its version is inconsistent.

    ``inline`` is intended to help achieve:

    - Composing existing ONNX models.
    - Interfacing with other ONNX libraries such as ``skl2onnx``.
    - Interface with custom operators.

    Parameters
    ----------
    model
        Target model to inline.

    Returns
    -------
    _InlineCall
        A callable which takes ``Var`` arguments and returns a
        dictionary of output names into ``Var``.

        Positional arguments are assigned based on the order they are
        listed in the model.  Keyword arguments are assigned based on
        their names in the model.

        Unspecified arguments are replaced by an initializer of the
        same name in the model, if one exists. This essentially
        uses them as a _default argument_.

        Input types are expected to be compatible with the model's
        graph input types.  Output types produced are copied from the
        model's graph output types.

    Raises
    ------
    TypeError
        If the arguments to the callback are supplied incorrectly or
        the variables are of the wrong type.

    Notes
    -----
    At build time, an inlined model puts its nodes at the assigned
    insertion point in the topological ordering.  Prefixing is applied
    (with the build system name) to attempt to avoid collisions.

    Initializers present in the model are replaced with Constant nodes,
    unless they are used as a default argument value. In this case
    they are also built as initializers.

    Symbolic dimensions in input and output shapes are stripped from
    the model and ignored, to avoid handling them inconsistently.

    Build behaviour should be treated as an implementation detail and
    may change.

    Currently, inlining models with subgraphs or functions is not supported
    as it cannot be performed in most cases due to lack of upstream support.
    """
    in_names = [i.name for i in model.graph.input]
    in_defaults = {i.name: i for i in model.graph.initializer}
    out_names = [o.name for o in model.graph.output]
    _defaults_msg = f" (defaults {list(in_defaults.keys())})"
    _signature_msg = f"signature {in_names}{_defaults_msg} -> {out_names}"

    model = _copy_model(model)
    # FIXME: Support for functions is a bit involved, as it interacts with build.
    if model.functions:
        raise ValueError(
            "Inlining models with functions is not supported. "
            "Consider removing or inlining the function definitions "
            "(if that preserves the model validity)."
        )
    # Handling symbolic dimensions is difficult and not particularly useful, so strip them
    for info in itertools.chain(
        model.graph.input, model.graph.output, model.graph.value_info
    ):
        info.type.CopyFrom(
            _strip_dim_symbol(Type._from_onnx(info.type), lambda x: True)._to_onnx()
        )
    # We handle everything related to initializers here, as currently build does not support them too well
    # Overridable initializers are saved to in_defaults, non-overridable replaced with Constant
    preamble: List[onnx.NodeProto] = []
    input_names = {i.name for i in model.graph.input}
    preamble.extend(
        onnx.helper.make_node("Constant", [], [i.name], value=i)
        for i in model.graph.initializer
        if i.name not in input_names
    )
    preamble.extend(
        onnx.helper.make_node("Constant", [], [i.values.name], sparse_value=i)
        for i in model.graph.sparse_initializer
        if i.values.name not in input_names
    )
    del model.graph.initializer[:]
    del model.graph.sparse_initializer[:]
    # The API on the protobuf list is a bit limited
    # - this prepends the preamble before the rest of the nodes
    model.graph.node.reverse()
    model.graph.node.extend(reversed(preamble))
    model.graph.node.reverse()
    # Now we can assume the graph has no initializers

    def inline_inner(*args: Var, **kwargs: Var) -> Dict[str, Var]:
        for name, arg in zip(in_names, args):
            if name in kwargs:
                raise TypeError(
                    f"inline callback got multiple values for argument '{name}', {_signature_msg}."
                )
            kwargs[name] = arg
        if not (missing := set(in_names) - set(kwargs)) <= set(in_defaults):
            raise TypeError(
                f"inline callback missing required arguments: {missing}, {_signature_msg}."
            )
        for name in missing:
            array = to_array(in_defaults[name])
            if array.dtype == np.dtype(object):
                array = array.astype(str)
            kwargs[name] = initializer(array)

        if set(kwargs) != set(in_names):
            raise TypeError(
                f"Error processing arguments, got {set(kwargs)}, expected {set(in_names)}."
            )
        node = _Inline(
            inputs=_Inline.Inputs([kwargs[name] for name in in_names]),
            out_variadic=len(model.graph.output),
            model=model,
        )
        return dict(zip(out_names, node.outputs.outputs))

    return inline_inner


__all__ = ["argument", "build", "inline"]



import numpy as np
import pytest

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._type_system import Tensor


@pytest.mark.parametrize("T,S", [(np.int64, np.str_), (np.str_, np.int64)])
def test_category_mapper_inference(T, S):
    (x,) = arguments(x=Tensor(T, (5, "N")))
    y = op_ml.category_mapper(x, cats_int64s=(0, 1, 2), cats_strings=("a", "b", "c"))
    assert y.type == Tensor(S, (5, "N"))



import pytest

import spox.opset.ai.onnx.v17 as op
from spox._graph import arguments
from spox._standard import InferenceError
from spox._type_system import Tensor


def test_compress_inference():
    x, y = arguments(x=Tensor(float, ("N", "M")), y=Tensor(bool, (None,)))
    assert op.compress(x, y).unwrap_tensor() == Tensor(float, (None,))
    assert op.compress(x, y, axis=0).unwrap_tensor() == Tensor(float, (None, "M"))
    assert op.compress(x, y, axis=1).unwrap_tensor() == Tensor(float, ("N", None))


def test_compress_inference_checks_bool_cond():
    (x,) = arguments(x=Tensor(float, ("N", "M")))
    with pytest.raises((InferenceError, RuntimeError), match="InferenceError"):
        op.compress(x, op.const(123)).unwrap_tensor()
    with pytest.raises((InferenceError, RuntimeError), match="InferenceError"):
        op.compress(x, op.const("abc")).unwrap_tensor()



import typing
from typing import Any, Callable, ClassVar, Optional, TypeVar, Union

import numpy

from . import _type_system, _value_prop

if typing.TYPE_CHECKING:
    from ._node import Node

F = TypeVar("F", bound=Callable)


class NotImplementedOperatorDispatcher:
    def _not_impl(self, *_):
        return NotImplemented

    add = sub = mul = truediv = floordiv = neg = and_ = or_ = xor = not_ = _not_impl


class Var:
    """
    Abstraction for a single ONNX value - like a tensor - that can be passed around in Python code.

    A ``Var`` represents some output of an operator.
    This operator is stored internally to allow reproducing the graph.

    The ``type`` field is inferred and checked by operators.
    It may be ``None`` if type inference failed, in which case it is unknown and should pass all type checks.
    However, untyped ``Var`` objects may not be used in some contexts.
    Keep in mind that the types themselves may have some information missing.
    For instance, tensors allow missing rank and shape information.

    There is an implicit value propagation mechanism, powered by the ONNX reference implementation.
    Values may be propagated if a ``Var`` always has a known and constant value at runtime.
    This is used for type & shape inference. For instance, Reshape to a constant shape can have the shape inferred.

    ``Var`` should be treated as strictly immutable.
    If a ``Var`` or any of its fields are modified, the behaviour is undefined and the produced graph may be invalid.

    Protected fields are to be treated as internal.
    Useful data is also shown by the string representation, but it should be treated as debug information.

    Should not be constructed directly - the main source of ``Var`` objects are operator constructors.
    """

    type: Optional[_type_system.Type]
    _value: Optional[_value_prop.PropValue]
    _op: "Node"
    _name: Optional[str]

    _operator_dispatcher: ClassVar[Any] = NotImplementedOperatorDispatcher()

    def __init__(
        self,
        op: "Node",
        type_: Optional[_type_system.Type],
        value: Optional[_value_prop.PropValue] = None,
    ):
        """The initializer of ``Var`` is protected. Use operator constructors to construct them instead."""
        if type_ is not None and not isinstance(type_, _type_system.Type):
            raise TypeError("The type field of a Var must be a Spox Type.")
        if value is not None and not isinstance(value, _value_prop.PropValue):
            raise TypeError("The propagated value field of a Var must be a PropValue.")
        if value is not None and value.type != type_:
            raise ValueError(
                f"The propagated value type ({value.type}) and actual Var type ({type_}) must be the same."
            )

        self.type = type_
        self._value = value
        self._op = op
        self._name = None

    def _rename(self, name: Optional[str]):
        """Mutates the internal state of the Var, overriding its name as given."""
        self._name = name

    @property
    def _which_output(self) -> Optional[str]:
        """Return the name of the output field that this var is stored in under ``self._op``."""
        if self._op is None:
            return None
        op_outs = self._op.outputs.get_vars()
        candidates = [key for key, var in op_outs.items() if var is self]
        return candidates[0] if candidates else None

    def _get_value(self) -> "_value_prop.ORTValue":
        """Get the propagated value in this Var and convert it to the ORT format. Raises if value is missing."""
        if self._value is None:
            raise ValueError("No propagated value associated with this Var.")
        return self._value.to_ort_value()

    def __repr__(self) -> str:
        nm = repr(self._name) + " " if self._name is not None else ""
        op_repr = self._op.get_op_repr() if self._op else "??"
        which = self._which_output
        is_unary = len(self._op.outputs) <= 1 if self._op else True
        which_repr = "->??" if which is None else (f"->{which}" if is_unary else "")
        return (
            f"<Var {nm}from {op_repr}{which_repr} of {self.type}"
            f"{'' if self._value is None else ' = ' + str(self._value.to_ort_value())}>"
        )

    def unwrap_type(self) -> _type_system.Type:
        """
        Return the :class:`~spox.Type` of ``self``, unless it is unknown.

        Returns
        -------
        _type_system.Type
            The type of the Var.
        Raises
        ------
        TypeError
            If ``type is None`` (the type of this ``Var`` is unknown).
        """
        if self.type is None:
            raise TypeError("Cannot unwrap requested type for Var, as it is unknown.")
        return self.type

    def unwrap_tensor(self) -> _type_system.Tensor:
        """Equivalent to ``self.unwrap_type().unwrap_tensor()``."""
        return self.unwrap_type().unwrap_tensor()

    def unwrap_sequence(self) -> _type_system.Sequence:
        """Equivalent to ``self.unwrap_type().unwrap_sequence()``."""
        return self.unwrap_type().unwrap_sequence()

    def unwrap_optional(self) -> _type_system.Optional:
        """Equivalent to ``self.unwrap_type().unwrap_optional()``."""
        return self.unwrap_type().unwrap_optional()

    def __copy__(self) -> "Var":
        # Simply return `self` to ensure that "copies" are still equal
        # during the build process
        return self

    def __deepcopy__(self, _) -> "Var":
        raise ValueError("'Var' objects cannot be deepcopied.")

    def __add__(self, other) -> "Var":
        return Var._operator_dispatcher.add(self, other)

    def __sub__(self, other) -> "Var":
        return Var._operator_dispatcher.sub(self, other)

    def __mul__(self, other) -> "Var":
        return Var._operator_dispatcher.mul(self, other)

    def __truediv__(self, other) -> "Var":
        return Var._operator_dispatcher.truediv(self, other)

    def __floordiv__(self, other) -> "Var":
        return Var._operator_dispatcher.floordiv(self, other)

    def __neg__(self) -> "Var":
        return Var._operator_dispatcher.neg(self)

    def __and__(self, other) -> "Var":
        return Var._operator_dispatcher.and_(self, other)

    def __or__(self, other) -> "Var":
        return Var._operator_dispatcher.or_(self, other)

    def __xor__(self, other) -> "Var":
        return Var._operator_dispatcher.xor(self, other)

    def __invert__(self) -> "Var":
        return Var._operator_dispatcher.not_(self)

    def __radd__(self, other) -> "Var":
        return Var._operator_dispatcher.add(other, self)

    def __rsub__(self, other) -> "Var":
        return Var._operator_dispatcher.sub(other, self)

    def __rmul__(self, other) -> "Var":
        return Var._operator_dispatcher.mul(other, self)

    def __rtruediv__(self, other) -> "Var":
        return Var._operator_dispatcher.truediv(other, self)

    def __rfloordiv__(self, other) -> "Var":
        return Var._operator_dispatcher.floordiv(other, self)

    def __rand__(self, other) -> "Var":
        return Var._operator_dispatcher.and_(other, self)

    def __ror__(self, other) -> "Var":
        return Var._operator_dispatcher.or_(other, self)

    def __rxor__(self, other) -> "Var":
        return Var._operator_dispatcher.xor(other, self)


def result_type(
    *types: Union[Var, numpy.generic, int, float],
) -> typing.Type[numpy.generic]:
    """Promote type for all given element types/values using ``np.result_type``."""
    return numpy.dtype(
        numpy.result_type(
            *(
                typ.unwrap_tensor().dtype if isinstance(typ, Var) else typ
                for typ in types
            )
        )
    ).type



import numpy as np

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments
from spox._type_system import Tensor


def test_tree_ensemble_regressor_inference_str():
    (x,) = arguments(x=Tensor(np.float32, ("N", 5)))
    y, z = op_ml.tree_ensemble_classifier(
        x, classlabels_strings=["a", "b"], class_ids=[1, 2, 3]
    )
    assert y.type == Tensor(np.str_, ("N",))
    assert z.type == Tensor(np.float32, ("N", 3))


def test_tree_ensemble_regressor_inference_int():
    (x,) = arguments(x=Tensor(np.float32, ("N", 5)))
    y, z = op_ml.tree_ensemble_classifier(
        x, classlabels_int64s=[1, 2], class_ids=[1, 2, 3]
    )
    assert y.type == Tensor(np.int64, ("N",))
    assert z.type == Tensor(np.float32, ("N", 3))



import onnx.shape_inference

InferenceError = onnx.shape_inference.InferenceError
ValidationError = onnx.checker.ValidationError


class InferenceWarning(Warning):
    """Warning related to partial typing of Variables.

    Incomplete type information may lead to reduced code safety or
    failure to build the model. The most common underlying cause for
    this warning is a missing or incomplete type inference for an
    operator in the upstream ``onnx`` project. Type and shape
    information may be manually reset by using (unsafe) casts or
    reshapes.
    """


class BuildError(Exception):
    """An error within the build process.

    Usually this means that Spox failed to resolve the graph structure
    due to an internal error - for example if the implicit Node graph
    was tampered with or the algorithm had a fault.
    """

    pass


__all__ = ["InferenceWarning", "InferenceError", "ValidationError"]



import numpy as np

import spox.opset.ai.onnx.ml.v3 as op_ml
from spox._graph import arguments, results
from spox._type_system import Tensor


def test_one_hot_encoder_inference():
    """Test type and shape inference for ``OneHotEncoder``."""
    (str_cats, int_cats) = arguments(
        str_cats=Tensor(np.str_, ("N", 5)), int_cats=Tensor(np.int64, ("N", 5))
    )
    str_encoded = op_ml.one_hot_encoder(str_cats, cats_strings=["a", "b", "c"])
    int_encoded = op_ml.one_hot_encoder(int_cats, cats_int64s=[0, 1, 2, 3])
    assert str_encoded.type == Tensor(np.float32, ("N", 5, 3))
    assert int_encoded.type == Tensor(np.float32, ("N", 5, 4))


def test_one_hot_encoder_runs(onnx_helper):
    """Test custom type and shape inference and correct runtime result of the ``OneHotEncoder``."""
    (str_cats,) = arguments(str_cats=Tensor(np.str_, ("N",)))
    encoded = op_ml.one_hot_encoder(str_cats, cats_strings=np.array(["a", "b", "c"]))
    graph = results(encoded=encoded)
    onnx_helper.assert_close(
        onnx_helper.run(graph, "encoded", str_cats=["a", "b", "c", "z"]),
        np.array(
            [
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0],
            ]
        ),
    )
